from enum import IntEnum, unique


@unique
class FrameType(IntEnum):
    """Types of Frames"""

    FRAME_CONFIG = 0
    FRAME_TEST = 0x1
    FRAME_WORK = 0x2
    FRAME_UNKNOWN = 0x3


class FrameHeader(IntEnum):
    """Frame headers"""

    CONFIG_TYPE1 = 0b0000
    """Type I, random seed"""
    CONFIG_TYPE2 = 0b0001
    """Type II, parameter REG"""
    CONFIG_TYPE3 = 0b0010
    """Type III, neuron RAM"""
    CONFIG_TYPE4 = 0b0011
    """Type IV, weight RAM"""

    TEST_TYPE1 = 0b0100
    TEST_TYPE2 = 0b0101
    TEST_TYPE3 = 0b0110
    TEST_TYPE4 = 0b0111

    WORK_TYPE1 = 0b1000
    """Type I, spike"""
    WORK_TYPE2 = 0b1001
    """Type II, sync"""
    WORK_TYPE3 = 0b1010
    """Type III, clear"""
    WORK_TYPE4 = 0b1011
    """Type IV, init"""


class FrameFormat:
    """General frame mask & offset."""

    GENERAL_MASK = (1 << 64) - 1

    # Header
    GENERAL_HEADER_OFFSET = 60
    GENERAL_HEADER_MASK = (1 << 4) - 1

    GENERAL_FRAME_TYPE_OFFSET = GENERAL_HEADER_OFFSET
    GENERAL_FRAME_TYPE_MASK = GENERAL_HEADER_MASK

    # Chip address
    GENERAL_CHIP_ADDR_OFFSET = 50
    GENERAL_CHIP_ADDR_MASK = (1 << 10) - 1
    # Chip X/Y address
    GENERAL_CHIP_X_ADDR_OFFSET = 55
    GENERAL_CHIP_X_ADDR_MASK = (1 << 5) - 1
    GENERAL_CHIP_Y_ADDR_OFFSET = GENERAL_CHIP_ADDR_OFFSET
    GENERAL_CHIP_Y_ADDR_MASK = (1 << 5) - 1

    # Core address
    GENERAL_CORE_ADDR_OFFSET = 40
    GENERAL_CORE_ADDR_MASK = (1 << 10) - 1
    # Core X/Y address
    GENERAL_CORE_X_ADDR_OFFSET = 45
    GENERAL_CORE_X_ADDR_MASK = (1 << 5) - 1
    GENERAL_CORE_Y_ADDR_OFFSET = GENERAL_CORE_ADDR_OFFSET
    GENERAL_CORE_Y_ADDR_MASK = (1 << 5) - 1

    # Core* address
    GENERAL_CORE_EX_ADDR_OFFSET = 30
    GENERAL_CORE_EX_ADDR_MASK = (1 << 10) - 1
    # Core* X/Y address
    GENERAL_CORE_X_EX_ADDR_OFFSET = 35
    GENERAL_CORE_X_EX_ADDR_MASK = (1 << 5) - 1
    GENERAL_CORE_Y_EX_ADDR_OFFSET = GENERAL_CORE_EX_ADDR_OFFSET
    GENERAL_CORE_Y_EX_ADDR_MASK = (1 << 5) - 1

    # Global core = Chip address + core address
    GENERAL_CORE_GLOBAL_ADDR_OFFSET = GENERAL_CORE_ADDR_OFFSET
    GENERAL_CORE_GLOBAL_ADDR_MASK = (1 << 20) - 1

    # Frame 前34位
    GENERAL_FRAME_PRE_OFFSET = 30
    GENERAL_FRAME_PRE_MASK = (1 << 34) - 1

    # 通用数据帧LOAD掩码
    GENERAL_PAYLOAD_OFFSET = 0
    GENERAL_PAYLOAD_MASK = (1 << 30) - 1

    # 通用数据包LOAD掩码
    GENERAL_PACKAGE_SRAM_ADDR_OFFSET = 20
    GENERAL_PACKAGE_SRAM_ADDR_MASK = (1 << 10) - 1

    GENERAL_PACKAGE_TYPE_OFFSET = 19
    GENERAL_PACKAGE_TYPE_MASK = 1

    GENERAL_PACKAGE_NUM_OFFSET = 0
    GENERAL_PACKAGE_NUM_MASK = (1 << 19) - 1


class OfflineFrameFormat(FrameFormat):
    """Basic offline frame format"""

    pass


class OnlineFrameFormat(FrameFormat):
    """Basic online frame format"""

    pass


class OfflineConfigFrame1Format(OfflineFrameFormat):
    """Offline config frame type I, random seed"""

    pass


class OfflineConfigFrame2Format(OfflineFrameFormat):
    """Offline config frame type II, parameter REG"""

    # Frame #1
    WEIGHT_WIDTH_OFFSET = 28
    WEIGHT_WIDTH_MASK = (1 << 2) - 1

    LCN_OFFSET = 24
    LCN_MASK = (1 << 4) - 1

    INPUT_WIDTH_OFFSET = 23
    INPUT_WIDTH_MASK = 1

    SPIKE_WIDTH_OFFSET = 22
    SPIKE_WIDTH_MASK = 1

    NEURON_NUM_OFFSET = 9
    NEURON_NUM_MASK = (1 << 13) - 1

    POOL_MAX_OFFSET = 8
    POOL_MAX_MASK = 1

    TICK_WAIT_START_HIGH8_OFFSET = 0
    TICK_WAIT_START_COMBINATION_OFFSET = 7
    TICK_WAIT_START_HIGH8_MASK = (1 << 8) - 1

    # Frame #2
    TICK_WAIT_START_LOW7_OFFSET = 23
    TICK_WAIT_START_LOW7_MASK = (1 << 7) - 1

    TICK_WAIT_END_OFFSET = 8
    TICK_WAIT_END_MASK = (1 << 15) - 1

    SNN_EN_OFFSET = 7
    SNN_EN_MASK = 1

    TARGET_LCN_OFFSET = 3
    TARGET_LCN_MASK = (1 << 4) - 1

    # 用于配置帧2型test_chip_addr
    TEST_CHIP_ADDR_HIGH3_OFFSET = 0
    TEST_CHIP_ADDR_LOW7_OFFSET = 7
    TEST_CHIP_ADDR_HIGH3_MASK = (1 << 3) - 1

    # Frame #3
    TEST_CHIP_ADDR_LOW7_OFFSET = 23
    TEST_CHIP_ADDR_LOW7_MASK = (1 << 7) - 1


class OfflineConfigFrame3Format(OfflineFrameFormat):
    """Offline config frame type III, param RAM"""

    # Frame #1
    VJT_PRE_OFFSET = 0
    VJT_PRE_MASK = (1 << 30) - 1

    BIT_TRUNCATE_OFFSET = 30
    BIT_TRUNCATE_MASK = (1 << 5) - 1

    WEIGHT_DET_STOCH_OFFSET = 35
    WEIGHT_DET_STOCH_MASK = (1 << 1) - 1

    LEAK_V_LOW28_OFFSET = 36
    LEAK_V_LOW28_MASK = (1 << 28) - 1

    # Frame #2
    LEAK_V_HIGH2_OFFSET = 0
    LEAK_V_HIGH2_MASK = (1 << 2) - 1

    LEAK_DET_STOCH_OFFSET = 2
    LEAK_DET_STOCH_MASK = (1 << 1) - 1

    LEAK_REVERSAL_FLAG_OFFSET = 3
    LEAK_REVERSAL_FLAG_MASK = (1 << 1) - 1

    THRESHOLD_POS_OFFSET = 4
    THRESHOLD_POS_MASK = (1 << 29) - 1

    THRESHOLD_NEG_OFFSET = 33
    THRESHOLD_NEG_MASK = (1 << 29) - 1

    THRESHOLD_NEG_MODE_OFFSET = 62
    THRESHOLD_NEG_MODE_MASK = (1 << 1) - 1

    THRESHOLD_MASK_CTRL_LOW1_OFFSET = 63
    THRESHOLD_MASK_CTRL_LOW1_MASK = (1 << 1) - 1

    # Frame #3
    THRESHOLD_MASK_CTRL_HIGH4_OFFSET = 0
    THRESHOLD_MASK_CTRL_HIGH4_MASK = (1 << 4) - 1

    LEAK_POST_OFFSET = 4
    LEAK_POST_MASK = (1 << 1) - 1

    RESET_V_OFFSET = 5
    RESET_V_MASK = (1 << 30) - 1

    RESET_MODE_OFFSET = 35
    RESET_MODE_MASK = (1 << 2) - 1

    ADDR_CHIP_Y_OFFSET = 37
    ADDR_CHIP_Y_MASK = (1 << 5) - 1

    ADDR_CHIP_X_OFFSET = 42
    ADDR_CHIP_X_MASK = (1 << 5) - 1

    ADDR_CORE_Y_EX_OFFSET = 47
    ADDR_CORE_Y_EX_MASK = (1 << 5) - 1

    ADDR_CORE_X_EX_OFFSET = 52
    ADDR_CORE_X_EX_MASK = (1 << 5) - 1

    ADDR_CORE_Y_OFFSET = 57
    ADDR_CORE_Y_MASK = (1 << 5) - 1

    ADDR_CORE_X_LOW2_OFFSET = 62
    ADDR_CORE_X_LOW2_MASK = (1 << 2) - 1

    # Frame #4
    ADDR_CORE_X_HIGH3_OFFSET = 0
    ADDR_CORE_X_HIGH3_MASK = (1 << 3) - 1

    ADDR_AXON_OFFSET = 3
    ADDR_AXON_MASK = (1 << 11) - 1

    TICK_RELATIVE_OFFSET = 14
    TICK_RELATIVE_MASK = (1 << 8) - 1


class OfflineConfigFrame4Format(OfflineFrameFormat):
    """Offline config frame type IV, weight RAM"""

    pass


RandomSeedFormat = OfflineConfigFrame1Format
ParameterRegFormat = OfflineConfigFrame2Format
ParameterRAMFormat = OfflineConfigFrame3Format
WeightRAMFormat = OfflineConfigFrame4Format


class OfflineWorkFrame1Format(OfflineFrameFormat):
    """Work frame type I"""

    RESERVED_OFFSET = 27
    RESERVED_MASK = (1 << 3) - 1

    AXON_OFFSET = 16
    AXON_MASK = (1 << 11) - 1

    TIMESLOT_OFFSET = 8
    TIMESLOT_MASK = (1 << 8) - 1

    DATA_OFFSET = 0
    DATA_MASK = (1 << 8) - 1


class OfflineWorkFrame2Format(OfflineFrameFormat):
    """Work frame type II"""

    RESERVED_OFFSET = 30
    RESERVED_MASK = (1 << 20) - 1

    TIME_OFFSET = 0
    TIME_MASK = (1 << 30) - 1


class OfflineWorkFrame3Format(OfflineFrameFormat):
    """Work frame type III"""

    RESERVED_OFFSET = 0
    RESERVED_MASK = (1 << 50) - 1


class OfflineWorkFrame4Format(OfflineFrameFormat):
    """Work frame type IV"""

    RESERVED_OFFSET = 0
    RESERVED_MASK = (1 << 50) - 1


SpikeFrameFormat = OfflineWorkFrame1Format
SyncFrameFormat = OfflineWorkFrame2Format
ClearFrameFormat = OfflineWorkFrame3Format
InitFrameFormat = OfflineWorkFrame4Format


# TODO frame format for online frames



from enum import Enum, IntEnum, unique
from typing import NamedTuple, Sequence, Set

from .coordinate import Coord
from .coordinate import ReplicationId as RId
from .hw_defs import HwParams

__all__ = [
    "RoutingLevel",
    "RoutingDirection",
    "RoutingStatus",
    "RoutingCost",
    "ROUTING_DIRECTIONS_IDX",
    "RoutingCoord",
    "get_routing_consumption",
    "get_multicast_cores",
    "get_replication_id",
]


@unique
class RoutingLevel(IntEnum):
    L0 = 0
    """Leaves of tree to store the data. A L0-cluster is a physical core."""
    L1 = 1
    L2 = 2
    L3 = 3
    L4 = 4
    L5 = 5


@unique
class RoutingDirection(Enum):
    """Indicate the 4 children of a cluster.

    NOTE: There is an X/Y coordinate priority method to specify the order.
    """

    X0Y0 = (0, 0)
    X0Y1 = (0, 1)
    X1Y0 = (1, 0)
    X1Y1 = (1, 1)
    ANY = (-1, -1)
    """Don't care when a level direction is `ANY`."""

    def to_index(self) -> int:
        """Convert the direction to index in children list."""
        if self is RoutingDirection.ANY:
            raise TypeError(f"The direction of routing is not specified.")

        x, y = self.value

        return (x << 1) + y


@unique
class RoutingStatus(IntEnum):
    """Indicate the status of L0-level cluster. Not used."""

    AVAILABLE = 0
    """Available for item to attach."""

    USED = 1
    """An item is attached to this cluster."""

    OCCUPIED = 2
    """Wasted. It will be an optimization goal."""

    ALL_EMPTY = 3
    """Not used."""


class RoutingCost(NamedTuple):
    n_L0: int
    n_L1: int
    n_L2: int
    n_L3: int
    n_L4: int

    def get_routing_level(self) -> RoutingLevel:
        """Return the routing cluster level.

        If the #N of Lx-level > 1, then we need a cluster with level Lx+1.
        And we need the #N of routing sub-level clusters.
        """
        for i in reversed(range(5)):
            if self[i] > 1:
                return RoutingLevel(i + 1)

        return RoutingLevel.L1


ROUTING_DIRECTIONS_IDX = (
    (
        RoutingDirection.X0Y0,
        RoutingDirection.X0Y1,
        RoutingDirection.X1Y0,
        RoutingDirection.X1Y1,
    )
    if HwParams.COORD_Y_PRIORITY
    else (
        RoutingDirection.X0Y0,
        RoutingDirection.X1Y0,
        RoutingDirection.X0Y1,
        RoutingDirection.X1Y1,
    )
)


class RoutingCoord(NamedTuple):
    """Use router directions to represent the coordinate of a cluster."""

    L4: RoutingDirection
    L3: RoutingDirection
    L2: RoutingDirection
    L1: RoutingDirection
    L0: RoutingDirection

    @property
    def level(self) -> RoutingLevel:
        for i in range(len(self)):
            if self[i] is RoutingDirection.ANY:
                return RoutingLevel(5 - i)

        return RoutingLevel.L0

    @property
    def coordinate(self) -> Coord:
        if self.level > RoutingLevel.L0:
            raise AttributeError("This property is only for L0-level cluster.")

        x = (
            (self.L4.value[0] << 4)
            + (self.L3.value[0] << 3)
            + (self.L2.value[0] << 2)
            + (self.L1.value[0] << 1)
            + self.L0.value[0]
        )

        y = (
            (self.L4.value[1] << 4)
            + (self.L3.value[1] << 3)
            + (self.L2.value[1] << 2)
            + (self.L1.value[1] << 1)
            + self.L0.value[1]
        )

        return Coord(x, y)


def get_routing_consumption(n_core: int) -> RoutingCost:
    """Get the consumption of clusters at different levels by given the `n_core`."""
    n_sub_node = HwParams.N_SUB_ROUTING_NODE

    # Find the nearest #N(=2^X) to accommodate `n_core` L0-level clusters.
    # If n_core = 5, return 8.
    # If n_core = 20, return 32.
    n_L0 = 1 << (n_core - 1).bit_length()
    n_L1 = 1 if n_L0 < n_sub_node else (n_L0 // n_sub_node)
    n_L2 = 1 if n_L1 < n_sub_node else (n_L1 // n_sub_node)
    n_L3 = 1 if n_L2 < n_sub_node else (n_L2 // n_sub_node)
    n_L4 = 1 if n_L3 < n_sub_node else (n_L3 // n_sub_node)

    return RoutingCost(n_L0, n_L1, n_L2, n_L3, n_L4)


def get_replication_id(coords: Sequence[Coord]) -> RId:
    """Get the replication ID by given the coordinates.

    Args:
        - coords: sequence of coordinates.
    """
    base_coord = coords[0]
    rid = RId(0, 0)

    for coord in coords[1:]:
        rid |= base_coord ^ coord

    return rid


def get_multicast_cores(base_coord: Coord, rid: RId) -> Set[Coord]:
    cores: Set[Coord] = set()
    corex = set()
    corey = set()
    temp = set()

    corex.add(base_coord.x)
    corey.add(base_coord.y)

    for lx in range(5):
        if (rid.x >> lx) & 1:
            for x in corex:
                temp.add(x ^ (1 << lx))

            corex = corex.union(temp)
            temp.clear()

        if (rid.y >> lx) & 1:
            for y in corey:
                temp.add(y ^ (1 << lx))

            corey = corey.union(temp)
            temp.clear()

    for x in corex:
        for y in corey:
            cores.add(Coord(x, y))

    return cores



import math
import sys
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import List, Sequence, Tuple, TypeVar, Union, final, overload

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from pydantic import Field
from pydantic.dataclasses import dataclass

from .hw_defs import HwParams

__all__ = [
    "Coord",
    "CoordOffset",
    "ReplicationId",
    "CoordLike",
    "RIdLike",
    "to_coord",
    "to_coordoffset",
    "to_rid",
]

CoordTuple: TypeAlias = Tuple[int, int]


def _xy_parser(other: Union[CoordTuple, "CoordOffset"]) -> CoordTuple:
    """Parse the coordinate in tuple format."""
    if not isinstance(other, (tuple, CoordOffset)):
        raise TypeError(f"Unsupported type: {type(other)}")

    if isinstance(other, tuple):
        if len(other) != 2:
            raise ValueError(f"Expect a tuple of 2 elements, but got {len(other)}.")

        return CoordOffset.from_tuple(other).to_tuple()
    else:
        return other.to_tuple()


class _CoordIdentifier(ABC):
    """Identifier to descripe coordinate of hardware unit. \
        The subclass of identifier must implement `__eq__` & `__ne__`."""

    @abstractmethod
    def __eq__(self, __other) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __ne__(self, __other) -> bool:
        raise NotImplementedError


@dataclass
class Coord(_CoordIdentifier):
    """Coordinate of the cores. Set coordinate (x, y) for every core.

    Left to right, +X, up to down, +Y.
    """

    x: int = Field(
        default=HwParams.CORE_X_MIN,
        ge=HwParams.CORE_X_MIN,
        le=HwParams.CORE_X_MAX,
    )
    y: int = Field(
        default=HwParams.CORE_Y_MIN,
        ge=HwParams.CORE_Y_MIN,
        le=HwParams.CORE_Y_MAX,
    )

    @classmethod
    def from_tuple(cls, pos: CoordTuple) -> "Coord":
        return cls(*pos)

    @classmethod
    def from_addr(cls, addr: int) -> "Coord":
        return cls(addr >> HwParams.N_BIT_CORE_Y, addr & HwParams.CORE_Y_MAX)

    def __add__(self, __other: "CoordOffset") -> "Coord":
        """
        Example:
        >>> c1 = Coord(1, 1)
        >>> c2 = c1 + CoordOffset(1, 1)
        >>> c1
        >>> Coord(2, 2)

        NOTE: `Coord` + `Coord` is meaningless.
        """
        if not isinstance(__other, CoordOffset):
            raise TypeError(f"Unsupported type: {type(__other)}")

        sum_x, sum_y = _sum_carry(self.x + __other.delta_x, self.y + __other.delta_y)

        return Coord(sum_x, sum_y)

    def __iadd__(self, __other: Union[CoordTuple, "CoordOffset"]) -> "Coord":
        """
        Example:
        >>> c1 = Coord(1, 1)
        >>> c1 += CoordOffset(1, 1)
        >>> c1
        >>> Coord(2, 2)
        """
        bx, by = _xy_parser(__other)
        self.x, self.y = _sum_carry(self.x + bx, self.y + by)

        return self

    @overload
    def __sub__(self, __other: "Coord") -> "CoordOffset":
        ...

    @overload
    def __sub__(self, __other: "CoordOffset") -> "Coord":
        ...

    def __sub__(
        self, __other: Union["Coord", "CoordOffset"]
    ) -> Union["Coord", "CoordOffset"]:
        """
        Example:
        >>> c1 = Coord(1, 1)
        >>> c2 = Coord(2, 2) - c1
        >>> c2
        >>> CoordOffset(1, 1)
        """
        if isinstance(__other, Coord):
            return CoordOffset(self.x - __other.x, self.y - __other.y)

        if isinstance(__other, CoordOffset):
            diff_x, diff_y = _sum_carry(
                self.x - __other.delta_x, self.y - __other.delta_y
            )
            return Coord(diff_x, diff_y)

        raise TypeError(f"Unsupported type: {type(__other)}")

    def __isub__(self, __other: Union[CoordTuple, "CoordOffset"]) -> "Coord":
        """
        Example:
        >>> c1 = Coord(2, 2)
        >>> c1 -= CoordOffset(1, 1)
        >>> c1
        >>> Coord(1, 1)
        """
        bx, by = _xy_parser(__other)
        self.x, self.y = _sum_carry(self.x - bx, self.y - by)

        return self

    """Operations below are used only when comparing with a Cooord."""

    def __eq__(self, __other: Union[CoordTuple, "Coord"]) -> bool:
        """
        Example:
        >>> Coord(4, 5) == Coord(4, 6)
        >>> False
        """
        if isinstance(__other, tuple):
            return self.to_tuple() == __other
        elif isinstance(__other, Coord):
            return self.x == __other.x and self.y == __other.y
        else:
            raise TypeError(f"Unsupported type: {type(__other)}")

    def __ne__(self, __other: Union[CoordTuple, "Coord"]) -> bool:
        return not self.__eq__(__other)

    # def __lt__(self, __other: "Coord") -> bool:
    #     """Whether the coord is on the left OR below of __other.

    #     Examples:
    #     >>> Coord(4, 5) < Coord(4, 6)
    #     True

    #     >>> Coord(4, 5) < Coord(5, 5)
    #     True

    #     >>> Coord(4, 5) < Coord(5, 3)
    #     True
    #     """
    #     if not isinstance(__other, Coord):
    #         raise TypeError(f"Unsupported type: {type(__other)}")

    #     return self.x < __other.x or self.y < __other.y

    # def __gt__(self, __other: "Coord") -> bool:
    #     """Whether the coord is on the right AND above of __other.

    #     Examples:
    #     >>> Coord(5, 5) > Coord(4, 5)
    #     True

    #     >>> Coord(4, 6) > Coord(4, 5)
    #     True

    #     >>> Coord(5, 4) > Coord(4, 5)
    #     False
    #     """
    #     if not isinstance(__other, Coord):
    #         raise TypeError(f"Unsupported type: {type(__other)}")

    #     # Except the `__eq__`
    #     return (
    #         (self.x > __other.x and self.y > __other.y)
    #         or (self.x == __other.x and self.y > __other.y)
    #         or (self.x > __other.x and self.y == __other.y)
    #     )

    # def __le__(self, __other: "Coord") -> bool:
    #     return self.__lt__(__other) or self.__eq__(__other)

    # def __ge__(self, __other: "Coord") -> bool:
    #     return self.__gt__(__other) or self.__eq__(__other)

    def __xor__(self, __other: "Coord") -> "ReplicationId":
        return ReplicationId(self.x ^ __other.x, self.y ^ __other.y)

    def __hash__(self) -> int:
        return hash(self.address)

    def __str__(self) -> str:
        return f"({self.x}, {self.y})"

    def __repr__(self) -> str:
        return f"Coord({self.x}, {self.y})"

    def to_tuple(self) -> CoordTuple:
        """Convert to tuple"""
        return (self.x, self.y)

    @property
    def address(self) -> int:
        """Convert to address, 10 bits"""
        return (self.x << HwParams.N_BIT_CORE_Y) | self.y


@final
class ReplicationId(Coord):
    def __and__(self, __other: Union[Coord, "ReplicationId"]) -> "ReplicationId":
        return ReplicationId(self.x & __other.x, self.y & __other.y)

    def __or__(self, __other: Union[Coord, "ReplicationId"]) -> "ReplicationId":
        return ReplicationId(self.x | __other.x, self.y | __other.y)

    def __xor__(self, __other: Union[Coord, "ReplicationId"]) -> "ReplicationId":
        return ReplicationId(self.x ^ __other.x, self.y ^ __other.y)

    # def __lshift__(self, __bit: int) -> int:
    #     return self.address << __bit

    # def __rshift__(self, __bit: int) -> int:
    #     return self.address >> __bit


class DistanceType(Enum):
    DISTANCE_ENCLIDEAN = auto()
    DISTANCE_MANHATTAN = auto()
    DISTANCE_CHEBYSHEV = auto()


@dataclass
class CoordOffset:
    """Offset of coordinate"""

    delta_x: int = Field(
        default=HwParams.CORE_X_MIN, ge=-HwParams.CORE_X_MAX, le=HwParams.CORE_X_MAX
    )
    delta_y: int = Field(
        default=HwParams.CORE_Y_MIN, ge=-HwParams.CORE_Y_MAX, le=HwParams.CORE_Y_MAX
    )

    @classmethod
    def from_tuple(cls, pos: CoordTuple) -> "CoordOffset":
        return cls(*pos)

    @overload
    def __add__(self, __other: Coord) -> Coord:
        ...

    @overload
    def __add__(self, __other: "CoordOffset") -> "CoordOffset":
        ...

    def __add__(
        self, __other: Union["CoordOffset", Coord]
    ) -> Union["CoordOffset", Coord]:
        """
        Examples:
        >>> delta_c1 = CoordOffset(1, 1)
        >>> delta_c2 = delta_c1 + CoordOffset(1, 1)
        >>> delta_c2
        >>> CoordOffset(2, 2)

        Coord = CoordOffset + Coord
        >>> delta_c = CoordOffset(1, 1)
        >>> c1 = Coord(2, 3)
        >>> c2 = delta_c + c1
        >>> c2
        >>> Coord(3, 4)
        """
        if isinstance(__other, CoordOffset):
            # Do not carry.
            return CoordOffset(
                self.delta_x + __other.delta_x, self.delta_y + __other.delta_y
            )
        elif isinstance(__other, Coord):
            sum_x, sum_y = _sum_carry(
                self.delta_x + __other.x, self.delta_y + __other.y
            )
            return Coord(sum_x, sum_y)
        else:
            raise TypeError(f"Unsupported type: {type(__other)}")

    def __iadd__(self, __other: Union[CoordTuple, "CoordOffset"]) -> "CoordOffset":
        """
        Example:
        >>> delta_c = CoordOffset(1, 1)
        >>> delta_c += CoordOffset(1, 1)
        >>> delta_c
        >>> CoordOffset(2, 2)
        """
        bx, by = _xy_parser(__other)
        self.delta_x += bx
        self.delta_y += by

        self._check()
        return self

    def __sub__(self, __other: "CoordOffset") -> "CoordOffset":
        """
        Example:
        >>> delta_c1 = CoordOffset(1, 1)
        >>> delta_c2 = CoordOffset(2, 2)
        >>> delta_c = delta_c1 - delta_c2
        >>> delta_c
        >>> CoordOffset(-1, -1)
        """
        if not isinstance(__other, CoordOffset):
            raise TypeError(f"Unsupported type: {type(__other)}")

        return CoordOffset(
            self.delta_x - __other.delta_x, self.delta_y - __other.delta_y
        )

    def __isub__(self, __other: Union[CoordTuple, "CoordOffset"]) -> "CoordOffset":
        """
        Example:
        >>> delta_c = CoordOffset(1, 1)
        >>> delta_c -= CoordOffset(1, 1)
        >>> delta_c
        >>> CoordOffset(0, 0)
        """
        bx, by = _xy_parser(__other)
        self.delta_x -= bx
        self.delta_y -= by

        self._check()
        return self

    def __eq__(self, __other: Union[CoordTuple, "CoordOffset"]) -> bool:
        """
        Example:
        >>> CoordOffset(4, 5) == CoordOffset(4, 6)
        >>> False
        """
        if isinstance(__other, tuple):
            return self.to_tuple() == __other
        elif isinstance(__other, CoordOffset):
            return self.delta_x == __other.delta_x and self.delta_y == __other.delta_y
        else:
            raise TypeError(f"Unsupported type: {type(__other)}")

    def __ne__(self, __other: "CoordOffset") -> bool:
        return not self.__eq__(__other)

    def to_tuple(self) -> CoordTuple:
        """Convert to tuple"""
        return (self.delta_x, self.delta_y)

    def to_distance(
        self, distance_type: DistanceType = DistanceType.DISTANCE_ENCLIDEAN
    ) -> Union[float, int]:
        """Distance between two coordinates."""
        if distance_type is DistanceType.DISTANCE_ENCLIDEAN:
            return self._euclidean_distance()
        elif distance_type is DistanceType.DISTANCE_MANHATTAN:
            return self._manhattan_distance()
        else:
            return self._chebyshev_distance()

    def _euclidean_distance(self) -> float:
        """Euclidean distance"""
        return math.sqrt(self.delta_x**2 + self.delta_y**2)

    def _manhattan_distance(self) -> int:
        """Manhattan distance"""
        return abs(self.delta_x) + abs(self.delta_y)

    def _chebyshev_distance(self) -> int:
        """Chebyshev distance"""
        return max(abs(self.delta_x), abs(self.delta_y))

    def _check(self) -> None:
        if (not -HwParams.CORE_X_MAX <= self.delta_x <= HwParams.CORE_X_MAX) or (
            not -HwParams.CORE_Y_MAX <= self.delta_y <= HwParams.CORE_Y_MAX
        ):
            raise ValueError(
                f"Offset of coordinate is out of range: ({self.delta_x}, {self.delta_y})."
            )


_x_crange = HwParams.CORE_X_MAX - HwParams.CORE_X_MIN + 1
_y_crange = HwParams.CORE_Y_MAX - HwParams.CORE_Y_MIN + 1


def _sum_carry(cx: int, cy: int) -> CoordTuple:
    if HwParams.COORD_Y_PRIORITY:
        if cx > HwParams.CORE_X_MAX:
            if cy < HwParams.CORE_Y_MAX:
                cx -= _x_crange
                cy += 1
            else:
                raise ValueError(
                    f"Coordinate of Y out of high limit: {HwParams.CORE_Y_MAX-1}({cy})."
                )
        elif cx < HwParams.CORE_X_MIN:
            if cy > HwParams.CORE_Y_MIN:
                cx += _x_crange
                cy -= 1
            else:
                raise ValueError(
                    f"Coordinate of Y out of low limit: {HwParams.CORE_Y_MIN}."
                )
    else:
        if cy > HwParams.CORE_Y_MAX:
            if cx < HwParams.CORE_X_MAX:
                cx += 1
                cy -= _y_crange
            else:
                raise ValueError(
                    f"Coordinate of X out of high limit: {HwParams.CORE_X_MAX-1}({cx})."
                )
        elif cy < HwParams.CORE_Y_MIN:
            if cx > HwParams.CORE_X_MIN:
                cx -= 1
                cy += _y_crange
            else:
                raise ValueError(
                    f"Coordinate of X out of low limit: {HwParams.CORE_X_MIN}."
                )

    return cx, cy


CoordLike = TypeVar("CoordLike", Coord, int, List[int], CoordTuple)
RIdLike = TypeVar("RIdLike", ReplicationId, int, List[int], CoordTuple)


def to_coord(coordlike: CoordLike) -> Coord:
    if isinstance(coordlike, int):
        return Coord.from_addr(coordlike)

    if isinstance(coordlike, (list, tuple)):
        if len(coordlike) != 2:
            raise TypeError(
                f"Expect a tuple or list of 2 elements, but got {len(coordlike)}."
            )

        return Coord(*coordlike)

    return coordlike


def to_coords(coordlikes: Sequence[CoordLike]) -> List[Coord]:
    return [to_coord(coordlike) for coordlike in coordlikes]


def to_coordoffset(offset: int) -> CoordOffset:
    return CoordOffset(
        offset % (HwParams.CORE_X_MAX + 1), offset // (HwParams.CORE_Y_MAX + 1)
    )


def to_rid(ridlike: RIdLike) -> ReplicationId:
    if isinstance(ridlike, int):
        return ReplicationId.from_addr(ridlike)

    if isinstance(ridlike, (list, tuple)):
        if len(ridlike) != 2:
            raise ValueError(
                f"Expect a tuple or list of 2 elements, but got {len(ridlike)}."
            )

        return ReplicationId(*ridlike)

    return ridlike



import sys
from typing import List, Tuple, TypeVar

import numpy as np
from numpy.typing import NDArray

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

FRAME_DTYPE: TypeAlias = np.uint64
FrameArrayType: TypeAlias = NDArray[FRAME_DTYPE]
ArrayType = TypeVar("ArrayType", List[int], Tuple[int, ...], np.ndarray)
BasicFrameArray = TypeVar(
    "BasicFrameArray", int, List[int], Tuple[int, ...], NDArray[FRAME_DTYPE]
)
IntScalarType = TypeVar("IntScalarType", int, np.integer)
DataType = TypeVar("DataType", int, np.integer, np.ndarray)
DataArrayType = TypeVar(
    "DataArrayType", int, np.integer, List[int], Tuple[int, ...], np.ndarray
)



from typing import Any, Dict, Union, overload
import numpy as np
from numpy.typing import NDArray

from paicorelib import Coord, LCN_EX, NeuronAttrs, NeuronDestInfo, ParamsReg
from paicorelib import ReplicationId as RId
from paicorelib import WeightPrecision as WP
from .frames import *
from ._types import (
    DataArrayType,
    FrameArrayType,
    FRAME_DTYPE,
    IntScalarType,
)


__all__ = ["OfflineFrameGen"]


class OfflineFrameGen:
    """Offline frame generator."""

    @staticmethod
    def gen_config_frame1(
        chip_coord: Coord, core_coord: Coord, rid: RId, /, random_seed: IntScalarType
    ) -> OfflineConfigFrame1:
        return OfflineConfigFrame1(chip_coord, core_coord, rid, int(random_seed))

    @overload
    @staticmethod
    def gen_config_frame2(
        chip_coord: Coord, core_coord: Coord, rid: RId, /, params_reg: ParamsReg
    ) -> OfflineConfigFrame2:
        ...

    @overload
    @staticmethod
    def gen_config_frame2(
        chip_coord: Coord, core_coord: Coord, rid: RId, /, params_reg: Dict[str, Any]
    ) -> OfflineConfigFrame2:
        ...

    @staticmethod
    def gen_config_frame2(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        params_reg: Union[ParamsReg, Dict[str, Any]],
    ) -> OfflineConfigFrame2:
        if isinstance(params_reg, ParamsReg):
            _params_reg = params_reg.model_dump(by_alias=True)
        else:
            _params_reg = params_reg

        return OfflineConfigFrame2(chip_coord, core_coord, rid, _params_reg)

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        n_neuron: IntScalarType,
        attrs: NeuronAttrs,
        dest_info: NeuronDestInfo,
        *,
        lcn_ex: LCN_EX = LCN_EX.LCN_1X,
        weight_precision: WP = WP.WEIGHT_WIDTH_1BIT,
    ) -> OfflineConfigFrame3:
        ...

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        n_neuron: IntScalarType,
        attrs: Dict[str, Any],
        dest_info: Dict[str, Any],
        *,
        lcn_ex: LCN_EX = LCN_EX.LCN_1X,
        weight_precision: WP = WP.WEIGHT_WIDTH_1BIT,
    ) -> OfflineConfigFrame3:
        ...

    @staticmethod
    def gen_config_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        n_neuron: IntScalarType,
        attrs: Union[NeuronAttrs, Dict[str, Any]],
        dest_info: Union[NeuronDestInfo, Dict[str, Any]],
        *,
        lcn_ex: LCN_EX = LCN_EX.LCN_1X,
        weight_precision: WP = WP.WEIGHT_WIDTH_1BIT,
    ) -> OfflineConfigFrame3:
        if isinstance(attrs, NeuronAttrs):
            _attrs = attrs.model_dump(by_alias=True)
        else:
            _attrs = attrs

        if isinstance(dest_info, NeuronDestInfo):
            _dest_info = dest_info.model_dump(by_alias=True)
        else:
            _dest_info = dest_info

        return OfflineConfigFrame3(
            chip_coord,
            core_coord,
            rid,
            int(sram_start_addr),
            int(n_neuron),
            _attrs,
            _dest_info,
            repeat=(1 << lcn_ex) * (1 << weight_precision),
        )

    @staticmethod
    def gen_config_frame4(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        n_data_package: IntScalarType,
        weight_ram: FrameArrayType,
    ) -> OfflineConfigFrame4:
        return OfflineConfigFrame4(
            chip_coord,
            core_coord,
            rid,
            int(sram_start_addr),
            int(n_data_package),
            weight_ram,
        )

    @staticmethod
    def gen_magic_init_frame(chip_coord: Coord, core_coord: Coord) -> FrameArrayType:
        """Magic initialization frames for PAICORE 2.0. DO NOT MODIFY!"""
        config1 = OfflineConfigFrame1(chip_coord, core_coord, RId(0, 0), 0)
        init_frame = OfflineWorkFrame4(chip_coord)
        work1 = OfflineWorkFrame1(chip_coord, core_coord, RId(0, 0), 0, 0, 0)

        v_config1 = config1.value
        return np.array(
            [
                v_config1[0],
                v_config1[1],
                init_frame.value[0],
                v_config1[2],
                work1.value[0],
            ],
            dtype=FRAME_DTYPE,
        )

    @staticmethod
    def gen_testin_frame1(
        chip_coord: Coord, core_coord: Coord, rid: RId, /
    ) -> OfflineTestInFrame1:
        return OfflineTestInFrame1(chip_coord, core_coord, rid)

    @staticmethod
    def gen_testout_frame1(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        random_seed: IntScalarType,
    ) -> OfflineTestOutFrame1:
        return OfflineTestOutFrame1(test_chip_coord, core_coord, rid, int(random_seed))

    @staticmethod
    def gen_testin_frame2(
        chip_coord: Coord, core_coord: Coord, rid: RId, /
    ) -> OfflineTestInFrame2:
        return OfflineTestInFrame2(chip_coord, core_coord, rid)

    @overload
    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord, core_coord: Coord, rid: RId, /, params_reg: ParamsReg
    ) -> OfflineTestOutFrame2:
        ...

    @overload
    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        params_reg: Dict[str, Any],
    ) -> OfflineTestOutFrame2:
        ...

    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        params_reg: Union[ParamsReg, Dict[str, Any]],
    ) -> OfflineTestOutFrame2:
        if isinstance(params_reg, ParamsReg):
            pr = params_reg.model_dump(by_alias=True)
        else:
            pr = params_reg

        return OfflineTestOutFrame2(test_chip_coord, core_coord, rid, pr)

    @staticmethod
    def gen_testin_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        data_package_num: IntScalarType,
    ) -> OfflineTestInFrame3:
        return OfflineTestInFrame3(
            chip_coord, core_coord, rid, int(sram_start_addr), int(data_package_num)
        )

    @staticmethod
    def gen_testout_frame3(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        n_neuron: IntScalarType,
        attrs: Union[NeuronAttrs, Dict[str, Any]],
        dest_info: Union[NeuronDestInfo, Dict[str, Any]],
        *,
        lcn_ex: LCN_EX = LCN_EX.LCN_1X,
        weight_precision: WP = WP.WEIGHT_WIDTH_1BIT,
    ) -> OfflineTestOutFrame3:
        if isinstance(attrs, NeuronAttrs):
            _attrs = attrs.model_dump(by_alias=True)
        else:
            _attrs = attrs

        if isinstance(dest_info, NeuronDestInfo):
            _dest_info = dest_info.model_dump(by_alias=True)
        else:
            _dest_info = dest_info

        return OfflineTestOutFrame3(
            chip_coord,
            core_coord,
            rid,
            int(sram_start_addr),
            int(n_neuron),
            _attrs,
            _dest_info,
            repeat=(1 << lcn_ex) * (1 << weight_precision),
        )

    @staticmethod
    def gen_testin_frame4(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        data_package_num: IntScalarType,
    ) -> OfflineTestInFrame4:
        return OfflineTestInFrame4(
            chip_coord, core_coord, rid, int(sram_start_addr), int(data_package_num)
        )

    @staticmethod
    def gen_testout_frame4(
        chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        sram_start_addr: IntScalarType,
        n_data_package: IntScalarType,
        weight_ram: FrameArrayType,
    ) -> OfflineTestOutFrame4:
        return OfflineTestOutFrame4(
            chip_coord,
            core_coord,
            rid,
            int(sram_start_addr),
            int(n_data_package),
            weight_ram,
        )

    @staticmethod
    def gen_work_frame1(
        one_input_node: Dict[str, Any],
        data: DataArrayType,
    ) -> FrameArrayType:
        """Generate the common part of the input spike frames by given the info \
            of one input node.

        Args:
            - one_input_node: a dictionary of one input node.
        """
        common_frame_dest = OfflineWorkFrame1._frame_dest_reorganized(one_input_node)
        _data = np.asarray(data, dtype=np.uint8)

        return OfflineFrameGen.gen_work_frame1_fast(common_frame_dest, _data)

    @staticmethod
    def gen_work_frame1_fast(
        frame_dest_info: FrameArrayType, data: NDArray[np.uint8]
    ) -> FrameArrayType:
        _max, _min = np.max(data, axis=None), np.min(data, axis=None)

        if _min < np.iinfo(np.uint8).min or _max > np.iinfo(np.uint8).max:
            raise ValueError(f"Data out of range int8 ({_min}, {_max}).")

        if frame_dest_info.size != data.size:
            raise ValueError(
                f"The size of frame dest info and data are not equal({frame_dest_info.size}, {data.size})."
            )

        return OfflineWorkFrame1._gen_frame_fast(frame_dest_info, data.flatten())

    @staticmethod
    def gen_work_frame2(chip_coord: Coord, n_sync: IntScalarType) -> OfflineWorkFrame2:
        return OfflineWorkFrame2(chip_coord, int(n_sync))

    @staticmethod
    def gen_work_frame3(chip_coord: Coord) -> OfflineWorkFrame3:
        return OfflineWorkFrame3(chip_coord)

    @staticmethod
    def gen_work_frame4(chip_coord: Coord) -> OfflineWorkFrame4:
        return OfflineWorkFrame4(chip_coord)



from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_serializer,
    model_validator,
)
from pydantic.type_adapter import TypeAdapter
from pydantic.types import NonNegativeInt
from typing_extensions import TypedDict  # Use `typing_extensions.TypedDict`.

from .coordinate import Coord
from .hw_defs import HwParams
from .reg_types import *

__all__ = ["CoreParams", "ParamsReg"]

WEIGHT_PRECISION_BIT_MAX = 2  # Not used
LCN_EXTENSION_BIT_MAX = 2  # Not used
INPUT_WIDTH_FORMAT_BIT_MAX = 2  # Not used
SPIKE_WIDTH_FORMAT_BIT_MAX = 2  # Not used
NUM_DENDRITE_BIT_MAX = 13
POOL_MAX_EN_BIT_MAX = 1  # Not used
TICK_WAIT_START_BIT_MAX = 15
TICK_WAIT_END_BIT_MAX = 15
SNN_MODE_EN_BIT_MAX = 1  # Not used
TARGET_LCN_BIT_MAX = 4
TEST_CHIP_ADDR_BIT_MAX = 10


class CoreParams(BaseModel):
    """Parameter model of register parameters listed in Section 2.4.1.

    NOTE: The parameters input in the model are declared in `docs/Table-of-Terms.md`.
    """

    model_config = ConfigDict(extra="ignore", validate_assignment=True)

    name: str = Field(description="Name of the physical core.", exclude=True)

    weight_precision: WeightPrecisionType = Field(
        le=WeightPrecisionType.WEIGHT_WIDTH_8BIT,
        serialization_alias="weight_width",
        description="Weight precision of crossbar.",
    )

    lcn_extension: LCNExtensionType = Field(
        le=LCNExtensionType.LCN_64X,
        serialization_alias="LCN",
        description="Scale of fan-in extension.",
    )

    input_width_format: InputWidthFormatType = Field(
        serialization_alias="input_width",
        description="Format of input spike.",
    )

    spike_width_format: SpikeWidthFormatType = Field(
        serialization_alias="spike_width",
        description="Format of output spike.",
    )

    num_dendrite: NonNegativeInt = Field(
        lt=(1 << NUM_DENDRITE_BIT_MAX),
        serialization_alias="neuron_num",
        description="The number of used dendrites.",
    )

    max_pooling_en: MaxPoolingEnableType = Field(
        default=MaxPoolingEnableType.DISABLE,
        serialization_alias="pool_max",
        description="Enable max pooling or not in 8-bit input format.",
    )

    tick_wait_start: NonNegativeInt = Field(
        default=0,
        lt=(1 << TICK_WAIT_START_BIT_MAX),
        description="The core begins to work at #N sync_all. 0 for not starting. Default is 0.",
    )

    tick_wait_end: NonNegativeInt = Field(
        default=0,
        lt=(1 << TICK_WAIT_END_BIT_MAX),
        description="The core keeps working within #N sync_all. 0 for not stopping. Default is 0.",
    )

    snn_mode_en: SNNModeEnableType = Field(
        default=SNNModeEnableType.ENABLE,
        serialization_alias="snn_en",
        description="Enable SNN mode or not.",
    )

    target_lcn: LCNExtensionType = Field(
        le=LCNExtensionType.LCN_64X,
        serialization_alias="target_LCN",
        description="LCN of the target core.",
    )

    test_chip_addr: Coord = Field(
        default=Coord(0, 0),
        description="Destination address of output test frames.",
    )

    """Parameter checks"""

    @model_validator(mode="after")
    def _neuron_num_range_limit(self):
        if self.input_width_format is InputWidthFormatType.WIDTH_1BIT:
            if self.num_dendrite > HwParams.N_DENDRITE_MAX_ANN:
                raise ValueError(
                    f"Param 'num_dendrite' out of range. When input width is 1-bit,"
                    f"The #N of dendrites should be no more than {HwParams.N_DENDRITE_MAX_ANN}."
                )
        else:
            if self.num_dendrite > HwParams.N_DENDRITE_MAX_SNN:
                raise ValueError(
                    f"Param 'num_dendrite' out of range. When input width is 8-bit,"
                    f"The #N of dendrite should be no more than {HwParams.N_DENDRITE_MAX_SNN}."
                )

        return self

    @model_validator(mode="after")
    def _max_pooling_en_check(self):
        if (
            self.input_width_format is InputWidthFormatType.WIDTH_1BIT
            and self.max_pooling_en is MaxPoolingEnableType.ENABLE
        ):
            self.max_pooling_en = MaxPoolingEnableType.DISABLE

        return self

    """Parameter serializers"""

    @field_serializer("weight_precision")
    def _weight_precision(self, weight_precision: WeightPrecisionType) -> int:
        return weight_precision.value

    @field_serializer("lcn_extension")
    def _lcn_extension(self, lcn_extension: LCNExtensionType) -> int:
        return lcn_extension.value

    @field_serializer("input_width_format")
    def _input_width_format(self, input_width_format: InputWidthFormatType) -> int:
        return input_width_format.value

    @field_serializer("spike_width_format")
    def _spike_width_format(self, spike_width_format: SpikeWidthFormatType) -> int:
        return spike_width_format.value

    @field_serializer("max_pooling_en")
    def _max_pooling_en(self, max_pooling_en: MaxPoolingEnableType) -> int:
        return max_pooling_en.value

    @field_serializer("snn_mode_en")
    def _snn_mode_en(self, snn_mode_en: SNNModeEnableType) -> int:
        return snn_mode_en.value

    @field_serializer("target_lcn")
    def _target_lcn(self, target_lcn: LCNExtensionType) -> int:
        return target_lcn.value

    @field_serializer("test_chip_addr")
    def _test_chip_addr(self, test_chip_addr: Coord) -> int:
        return test_chip_addr.address


ParamsReg = CoreParams


class _ParamsRegDict(TypedDict):
    """Typed dictionary of `ParamsReg` for typing check."""

    weight_width: int
    LCN: int
    input_width: int
    spike_width: int
    neuron_num: int
    pool_max: int
    tick_wait_start: int
    tick_wait_end: int
    snn_en: int
    target_LCN: int
    test_chip_addr: int


ParamsRegChecker = TypeAdapter(_ParamsRegDict)



from abc import ABC, abstractmethod
from typing import ClassVar, List, NamedTuple, Tuple

from .hw_defs import HwParams
from .reg_types import CoreMode

__all__ = ["NeuronSegment", "AxonCoord", "AxonSegment", "HwCore"]


class NeuronSegment(NamedTuple):
    """Segment of neuron.

    Example:
        index = (0, 100, 1)
        addr_offset = 10
        interval = 2
        addr_ram = (10, 11, ..., 210)

        for addr in addr_ram:
            write in addr with tick_relative and addr_axon at 'addr/interval'.
    """

    index: slice
    """The original index of this segment of neurons."""
    addr_offset: int
    """The offset of the RAM address."""
    interval: int = 1
    """The interval of address when mapping neuron attributes on the RAM."""

    @property
    def n_neuron(self) -> int:
        return self.index.stop - self.index.start

    @property
    def addr_max(self) -> int:
        if (
            _addr_max := self.addr_offset
            + self.interval * (self.index.stop - self.index.start)
        ) > HwParams.ADDR_RAM_MAX:
            raise ValueError(f"RAM Address out of {HwParams.ADDR_RAM_MAX}: {_addr_max}.")

        return _addr_max

    @property
    def addr_ram(self) -> List[int]:
        """Convert index of neuron into RAM address."""
        return list(range(self.addr_offset, self.addr_max, 1))

    @property
    def addr_slice(self) -> slice:
        """Display the RAM address in slice format."""
        return slice(
            self.addr_offset,
            self.addr_max,
            self.interval,
        )


class AxonCoord(NamedTuple):
    tick_relative: int
    addr_axon: int


class AxonSegment(NamedTuple):
    n_axon: int
    """#N of axons."""
    addr_width: int
    """The range of axon address is [addr_offset, addr_offset + addr_width)."""
    addr_offset: int
    """The offset of the assigned address."""


class HwCore(ABC):
    """Hardware core abstraction."""

    RUNTIME_MODE: ClassVar[CoreMode]

    @property
    @abstractmethod
    def shape(self) -> Tuple[int, int]:
        """Shape of the core."""
        raise NotImplementedError

    @property
    @abstractmethod
    def n_core_required(self) -> int:
        """#N of cores required to accommodate neurons inside self."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def build(cls):
        ...



from enum import Enum, IntEnum, unique

__all__ = [
    "ResetMode",
    "LeakingComparisonMode",
    "NegativeThresholdMode",
    "LeakingDirectionMode",
    "LeakingIntegrationMode",
    "SynapticIntegrationMode",
    "ThresholdMode",
]

"""
    Type defines of RAM parameters.
    See Section 2.4.2 in V2.1 Manual for details.
"""


@unique
class ResetMode(Enum):
    """Reset modes of neurons. 2-bit.

    - `MODE_NORMAL`: normal mode. Default value.
    - `MODE_LINEAR`: linear mode.
    - `MODE_NONRESET`: non-reset mode.

    NOTE: Same as `reset_mode` in V2.1.
    """

    MODE_NORMAL = 0  # Default value.
    MODE_LINEAR = 1
    MODE_NONRESET = 2


@unique
class LeakingComparisonMode(Enum):
    """Leak after comparison or before. 1-bit.

    - `LEAK_BEFORE_COMP`: leak before comparison.
    - `LEAK_AFTER_COMP`: leak after comparison. Default value.

    NOTE: Same as `leak_post` in V2.1.
    """

    LEAK_BEFORE_COMP = 0
    LEAK_AFTER_COMP = 1  # Default value.


@unique
class NegativeThresholdMode(Enum):
    """Modes of negative threshold. 1-bit.

    - `MODE_RESET`: reset mode. Default value.
    - `MODE_SATURATION`: saturation(floor) mode.

    NOTE: Same as `threshold_neg_mode` in V2.1.
    """

    MODE_RESET = 0  # Default value.
    MODE_SATURATION = 1


@unique
class LeakingDirectionMode(Enum):
    """Direction of leaking, forward or reversal.

    - `MODE_FORWARD`: forward leaking. Default value.
    - `MODE_REVERSAL`: reversal leaking.

    NOTE: Same as `leak_reversal_flag` in V2.1.
    """

    MODE_FORWARD = 0  # Default value.
    MODE_REVERSAL = 1


@unique
class LeakingIntegrationMode(Enum):
    """Mode of leaking integration, deterministic or stochastic.

    - `MODE_DETERMINISTIC`: deterministic leaking. Default value.
    - `MODE_STOCHASTIC`: stochastic leaking.

    NOTE: Same as `leak_det_stoch` in V2.1.
    """

    MODE_DETERMINISTIC = 0  # Default value.
    MODE_STOCHASTIC = 1


@unique
class SynapticIntegrationMode(Enum):
    """Modes of synaptic integration, deterministic or stochastic.

    - `MODE_DETERMINISTIC`: deterministic weights. Default value.
    - `MODE_STOCHASTIC`: stochastic weights.

    NOTE: Same as `weight_det_stoch` in V2.1.
    """

    MODE_DETERMINISTIC = 0  # Default value.
    MODE_STOCHASTIC = 1


@unique
class ThresholdMode(IntEnum):
    """Indicates whether the neuron reaches the threshold or not.

    - `NOT_EXCEEDED`: dosen't exceed. Must reset after neuronal reset.
    - `EXCEED_POSITIVE`: exceeded positive threshold.
    - `EXCEED_NEGATIVE`: exceeded negative threshold.
    """

    NOT_EXCEEDED = 0
    EXCEED_POSITIVE = 1
    EXCEED_NEGATIVE = 2
